﻿using System;


namespace Hardcodet.Commons.Diagnostics
{
  /// <summary>
  /// A guard that can be used to measure the time
  /// used for a given process by simply wrapping it
  /// in a using statement.<br/>
  /// Timestamp determination relies on the <see cref="SystemTime"/>
  /// helper class.<br/>
  /// Blog post:
  /// http://www.hardcodet.net/2009/02/timeguard-helper-class-to-measure-elapsed-time
  /// </summary>
  public class TimeGuard : IDisposable
  {
    /// <summary>
    /// An action delegate that is used to submit the
    /// determined timestamp.
    /// </summary>
    private readonly Action<double> action;

    /// <summary>
    /// The intial timestamp according to <see cref="SystemTime.Now"/>.
    /// </summary>
    private readonly DateTime startTime;


    /// <summary>
    /// Creates the guard with a given action that receives the
    /// elapsed time in milliseconds.
    /// </summary>
    /// <param name="setDelta">The action that is being invoked
    /// with time (in milleconds) that was taken between construction
    /// of this instance until disposal.</param>
    /// <exception cref="ArgumentNullException">If <paramref name="setDelta"/>
    /// is a null reference.</exception>
    public TimeGuard(Action<double> setDelta)
    {
      if (setDelta == null) throw new ArgumentNullException("setDelta");

      action = setDelta;
      startTime = SystemTime.Now();
    }


    /// <summary>
    /// Calculates the elapsed time since construction
    /// and submits the delta to the <see cref="action"/>
    /// action.
    /// </summary>
    public void Dispose()
    {
      var now = SystemTime.Now();
      action(now.Subtract(startTime).TotalMilliseconds);
    }

  }
}