﻿using System;
using System.Linq.Expressions;

namespace Hardcodet.Commons
{
  /// <summary>
  /// Provides common runtime validation functionality.
  /// </summary>
  public static class Ensure
  {
    /// <summary>
    /// Makes sure a given argument is not null.
    /// </summary>
    /// <typeparam name="T">Type of the argument.</typeparam>
    /// <param name="argument">The submitted parameter value.</param>
    /// <param name="argumentName">The name of the argument.</param>
    /// <exception cref="ArgumentNullException">If <paramref name="argument"/>
    /// is a null reference.</exception>
    public static void ArgumentNotNull<T>(T argument, string argumentName) where T : class
    {
      if (argument == null) throw new ArgumentNullException(argumentName);
    }


    /// <summary>
    /// Makes sure a given argument is not null.
    /// </summary>
    /// <typeparam name="T">Type of the argument.</typeparam>
    /// <param name="expression">A <see cref="MemberExpression"/> that refers
    /// to the parameter to be validated.</param>
    /// <exception cref="ArgumentNullException">If <paramref name="expression"/>
    /// links to a null reference.</exception>
    /// <remarks>Only use this method for validation scenarios where performance is
    /// not a critical issue. Otherwise revert to the <see cref="ArgumentNotNull{T}(T,string)"/>
    /// method that performs better by magnitudes.</remarks>
    public static void ArgumentNotNull<T>(Expression<Func<T>> expression) where T : class
    {
      //process the expression in order to get parameter name and value
      var pe = (MemberExpression) expression.Body;
      var paramterName = pe.Member.Name;
      var func = expression.Compile();

      //delegate validation to overload
      ArgumentNotNull(func(), paramterName);
    }
  }
}