﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using System.Threading;

namespace Hardcodet.Commons
{
  /// <summary>
  /// Global helper extensions.
  /// </summary>
  public static class Extensions
  {
    /// <summary>
    /// Checks a list of candidates for equality to a given
    /// reference value.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    /// <param name="value">The evaluated value.</param>
    /// <param name="candidates">A liste of possible values that are
    /// regarded valid.</param>
    /// <returns>True if one of the submitted <paramref name="candidates"/>
    /// matches the evaluated value. If the <paramref name="candidates"/>
    /// parameter itself is null, too, the method returns false as well,
    /// which allows to check with null values, too.</returns>
    /// <exception cref="ArgumentNullException">If <paramref name="candidates"/>
    /// is a null reference.</exception>
    public static bool Is<T>(this T value, params T[] candidates)
    {
      if (candidates == null) return false;

      foreach (var t in candidates)
      {
        if (value.Equals(t)) return true;
      }

      return false;
    }
    
    
    #region type evaluation
    
    /// <summary>
    /// Mimics the <c>is</c> keyword in order to check
    /// type compatibility based on types rather than
    /// static class expressions.
    /// </summary>
    /// <param name="obj">The object that is being evaluated.</param>
    /// <param name="parent">The type to compare to.</param>
    /// <returns>True if <paramref name="obj"/> can be declared as
    /// an instance of <paramref name="parent"/>, which is the case
    /// if its either of the same type, implements, or derives from
    /// <paramref name="parent"/>.</returns>
    public static bool IsA(this object obj, Type parent)
    {
      if (obj == null) throw new ArgumentNullException("obj");
      if (parent == null) throw new ArgumentNullException("parent");

      return IsA(obj.GetType(), parent);
    }


    /// <summary>
    /// Mimics the <c>is</c> keyword in order to check
    /// type compatibility based on types rather than
    /// static class expressions.
    /// </summary>
    /// <typeparam name="T">The type to compare to.</typeparam>
    /// <param name="type">The type that is being evaluated.</param>
    /// <returns>True if an instance of <paramref name="type"/> can
    /// be declared as an instance of <typeparamref name="T"/>,
    /// which is the case if its either of the same type, implements,
    /// or derives from <typeparamref name="T"/>.</returns>
    public static bool IsA<T>(this Type type)
    {
      return type.IsA(typeof (T));
    }


    /// <summary>
    /// Mimics the <c>is</c> keyword in order to check
    /// type compatibility based on types rather than
    /// static class expressions.
    /// </summary>
    /// <param name="type">The object that is being evaluated.</param>
    /// <param name="parent">The type to compare to.</param>
    /// <returns>True if an instance of <paramref name="type"/> can be declared as
    /// an instance of <paramref name="parent"/>, which is the case
    /// if its either of the same type, implements, or derives from
    /// <paramref name="parent"/>.</returns>
    public static bool IsA(this Type type, Type parent)
    {
      if (type == null) throw new ArgumentNullException("type");
      if (parent == null) throw new ArgumentNullException("parent");

      return parent.IsAssignableFrom(type);
    }
    
    #endregion

    #region collections

    /// <summary>
    /// Checks whether a given collection contains an item that matches
    /// the submitted predicate.
    /// </summary>
    /// <typeparam name="T">The collection's content.</typeparam>
    /// <param name="source">The sequence to be analyzed.</param>
    /// <param name="predicate">A function that tests each element for a
    /// given condition.</param>
    /// <returns>True if an item was found that matches the submitted
    /// <paramref name="predicate"/> function.</returns>
    /// <exception cref="ArgumentNullException">If <paramref name="predicate"/>
    /// is a null reference.</exception>
    public static bool Contains<T>(this IEnumerable<T> source, Func<T, bool> predicate) where T : class
    {
      if (predicate == null) throw new ArgumentNullException("predicate");
      return source.FirstOrDefault(predicate) != null;
    }


    /// <summary>
    /// Checks whether a collection contains an instance of a given type.
    /// </summary>
    /// <typeparam name="T">The collection's content.</typeparam>
    /// <param name="source">The sequence to be analyzed.</param>
    /// <param name="item">An instance to check for.</param>
    /// <returns>True if the submitted reference was found in the
    /// sequence.</returns>
    /// <exception cref="ArgumentNullException">If <paramref name="item"/>
    /// is a null reference.</exception>
    public static bool ContainsInstance<T>(this IEnumerable<T> source, T item) where T : class
    {
      if (item == null) throw new ArgumentNullException("item");

      return source.FirstOrDefault(i => ReferenceEquals(i, item)) != null;
    }


    /// <summary>
    /// Performs a given action on every item of a given sequence.
    /// </summary>
    /// <typeparam name="T">The collection's content.</typeparam>
    /// <param name="source">The sequence to be processed.</param>
    /// <param name="action">An action delegate that is being invoked
    /// for every item of the <paramref name="source"/> sequence.</param>
    /// <exception cref="ArgumentNullException">If <paramref name="action"/>
    /// is a null reference.</exception>
    public static void Do<T>(this IEnumerable<T> source, Action<T> action)
    {
      if (action == null) throw new ArgumentNullException("action");

      foreach (var item in source)
      {
        action(item);
      }
    }


    /// <summary>
    /// Triggers asynchronous invocation of a given action on every
    /// item of a given sequence using <see cref="ThreadPool"/>.
    /// </summary>
    /// <typeparam name="T">The collection's content.</typeparam>
    /// <param name="source">The sequence to be processed.</param>
    /// <param name="action">An action delegate that is being invoked
    /// for every item of the <paramref name="source"/> sequence.</param>
    /// <exception cref="ArgumentNullException">If <paramref name="action"/>
    /// is a null reference.</exception>
    public static void DoAsync<T>(this IEnumerable<T> source, Action<T> action)
    {
      if (action == null) throw new ArgumentNullException("action");

      foreach (var item in source)
      {
        ThreadPool.QueueUserWorkItem(state => action((T) state), item);
      }
    }

    #endregion

    #region string formatting

    /// <summary>
    /// Provides a shortcut to <see cref="String.Format(string,object[])"/>
    /// that allows to combine a string mask with a set or arguments.
    /// </summary>
    /// <param name="mask">The format item in which the arguments are
    /// being injected.</param>
    /// <param name="args">Arguments to be moved into the mask.</param>
    /// <returns>The formatted string.</returns>
    public static string Put(this string mask, params object[] args)
    {
      return String.Format(mask, args);
    }


    /// <summary>
    /// Writes an unformatted string to the console output.
    /// </summary>
    /// <param name="value">The string to write to the console.</param>
    public static void ToConsole(this string value)
    {
      Console.Out.WriteLine(value);
    }


    /// <summary>
    /// Formats a given string mask with a number of arguments and writes
    /// the formatted string to the console output.
    /// </summary>
    /// <param name="mask">The format item that in which the arguments are
    /// being injected.</param>
    /// <param name="args">Arguments to be moved into the mask.</param>
    /// <exception cref="ArgumentNullException">If <paramref name="mask"/>
    /// is a null reference.</exception>
    public static void ToConsole(this string mask, params object[] args)
    {
      if (mask == null) throw new ArgumentNullException("mask");
      Console.Out.WriteLine(mask, args);
    }


    /// <summary>
    /// Writes a given enum flag to the console.
    /// </summary>
    /// <param name="flag">The enum value to be written to the console.</param>
    public static void ToConsole(this Enum flag)
    {
      string enumType = flag.GetType().Name;
      Console.WriteLine("{0}: {1}", enumType, flag);
    }

    #endregion

    #region enums

    /// <summary>
    /// Resolves the value of a <see cref="DescriptionAttribute"/> that
    /// was used to decorate a given enum.
    /// </summary>
    /// <param name="value">^The enum value to be resolved.</param>
    /// <returns>The value of the description attribute that was used to
    /// decorate the enum. If no matching description is found, a string
    /// representation of the enum value is returned. If the enum value
    /// itself is null (nullable enum), null is returned.</returns>
    public static string GetDescription(this Enum value)
    {
      if (value == null) return null;

      // Get the Description attribute value for the enum value
      FieldInfo fi = value.GetType().GetField(value.ToString());

      var attributes = (DescriptionAttribute[])fi.GetCustomAttributes(
              typeof(DescriptionAttribute), false);

      if (attributes.Length > 0) return attributes[0].Description;
      return value.ToString();
    }

    #endregion

    #region events

    /// <summary>
    /// Raises an simple event that provides no data.
    /// </summary>
    public static void Raise(this EventHandler eventHandler, object sender)
    {
      var handler = eventHandler;
      if (handler != null) handler(sender, EventArgs.Empty);
    }


    /// <summary>
    /// Raises an event and submits the corresponding event arguments.
    /// </summary>
    public static void Raise<T>(this EventHandler<T> eventHandler, object sender, T eventArgs) where T:EventArgs
    {
      var handler = eventHandler;
      if (handler != null) handler(sender, eventArgs);
    }

    #endregion
    
  }
}