using System;
using System.Diagnostics;
using System.Globalization;
using System.Text;

namespace Hardcodet.Commons.Globalization
{
  /// <summary>
  /// Provides helper methods to deal with
  /// date / time values.
  /// </summary>
  [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "Util"), DebuggerNonUserCode]
  public static class DateTimeUtil
  {
    #region format constants

    /// <summary>
    /// Standard time pattern to be used throughout the application
    /// for displaying time strings that display full resolution
    /// timestamps (down to single ticks). 
    /// </summary>
    public const string MaxResolutionTimePattern = "HH:mm:ss.fffffff";

    /// <summary>
    /// Standard time pattern to be used throughout the application
    /// for displaying time strings that go down to hundreds of
    /// microseconds (0.1 ms).
    /// </summary>
    public const string HighResolutionTimePattern = "HH:mm:ss.ffff";

    /// <summary>
    /// Standard time pattern to be used throughout the application
    /// for displaying time strings that include milliseconds.
    /// </summary>
    public const string DetailedTimePattern = "HH:mm:ss.fff";

    /// <summary>
    /// Standardized, locale-independent time pattern that can be 
    /// used throughout an application for displaying short time strings.
    /// </summary>
    public const string ShortTimePattern = "HH:mm:ss";

    /// <summary>
    /// Gets a locale-specific date pattern which displays
    /// both date and time.
    /// </summary>
    public const string DateTimePattern = "G";

    #endregion
    

    #region (de)serialize timestamp

    /// <summary>
    /// Creates a human-readable string that represents a given
    /// timestamp in a format that can be converted back to the
    /// correct value across time zones.<br/>
    /// The serialized timestamp can be converted back into a
    /// <see cref="DateTime"/> instance through the
    /// <see cref="ReadDateTime"/> method.
    /// </summary>
    /// <param name="value">Timestamp to be serialized.</param>
    /// <returns>String representation of the timestamp.</returns>
    public static string SerializeDateTime(DateTime value)
    {
      return value.ToString("o", CultureInfo.InvariantCulture);
    }


    /// <summary>
    /// Parses a serialized timestamp that was created through the <see cref="SerializeDateTime"/>
    /// method.
    /// </summary>
    /// <param name="value">A string that represents a given time.</param>
    /// <returns>A corresponding timestamp.</returns>
    /// <remarks>This method works with local, UTC and unspecified timestamps.</remarks>
    public static DateTime ReadDateTime(string value)
    {
      return DateTime.Parse(value, CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind);
    }

    #endregion


    #region format / parse (short) date string

    /// <summary>
    /// Convenience method which formats a given date with
    /// the <see cref="DateTimeFormatInfo.ShortDatePattern"/> constant using a
    /// <c>DateTimeFormatInfo</c> that matches the requested
    /// locale.
    /// </summary>
    /// <param name="dt">The date/time value to be formatted.</param>
    /// <param name="locale">The targeted locale.</param>
    /// <returns>A formatted date string (short format).</returns>
    public static string FormatDateString(DateTime dt, TargetLocale locale)
    {
      DateTimeFormatInfo format = LocalizationUtil.GetDateFormat(locale);
      return dt.ToString(format.ShortDatePattern, format);
    }


    /// <summary>
    /// Reads a date / time string which was created through the
    /// <see cref="FormatDateString"/> method (relies on
    /// <see cref="DateTimeFormatInfo.ShortDatePattern"/>).
    /// </summary>
    /// <param name="value">A string that represents a given time.</param>
    /// <param name="locale">The targeted locale.</param>
    /// <returns>A DateTime object that represents the serialized time.</returns>
    /// <remarks>In order to ensure exact parsing, it is recommended to use the 
    /// <see cref="SerializeDateTime"/> and <see cref="ReadDateTime"/> methods.
    /// Use this method if a user friendly presentation is more important.
    /// </remarks>
    public static DateTime ParseFormattedDateString(string value, TargetLocale locale)
    {
      DateTimeFormatInfo format = LocalizationUtil.GetDateFormat(locale);
      string pattern = format.ShortDatePattern;

      return DateTime.ParseExact(value, pattern, format);
    }


    #endregion


    #region format parse (short) date-time string

    /// <summary>
    /// Convenience method which formats a given date / time with
    /// the <see cref="DateTimePattern"/> constant using a
    /// <c>DateTimeFormatInfo</c> that matches the requested
    /// locale.
    /// </summary>
    /// <param name="dt">The date/time value to be formatted.</param>
    /// <param name="locale">The targeted locale.</param>
    /// <returns>A formatted date string (short format).</returns>
    public static string FormatDateTimeString(DateTime dt, TargetLocale locale)
    {
      DateTimeFormatInfo format = LocalizationUtil.GetDateFormat(locale);
      return dt.ToString(DateTimePattern, format);
    }


    /// <summary>
    /// Reads a date / time string which was created through the
    /// <see cref="FormatDateTimeString"/> method (relying on the
    /// <see cref="DateTimePattern"/>).
    /// </summary>
    /// <param name="value">A string that represents a given time.</param>
    /// <param name="locale">The targeted locale.</param>
    /// <returns>A DateTime object that represents the serialized time.</returns>
    /// <remarks>In order to ensure exact parsing, it is recommended to use the 
    /// <see cref="SerializeDateTime"/> and <see cref="ReadDateTime"/> methods.
    /// Use this method if a user friendly presentation is more important.
    /// </remarks>
    public static DateTime ParseFormattedDateTimeString(string value, TargetLocale locale)
    {
      DateTimeFormatInfo format = LocalizationUtil.GetDateFormat(locale);
      const string pattern = DateTimePattern;

      return DateTime.ParseExact(value, pattern, format);
    }

    #endregion


    #region get short patterns

    /// <summary>
    /// Gets the short date pattern for a given locale which can
    /// be combined with other patterns (e.g. for formatting and
    /// parsing via the <see cref="FormatTimestamp"/> and
    /// <see cref="ParseTimestamp"/> methods.
    /// </summary>
    /// <param name="locale">The targeted locale.</param>
    /// <returns>The date pattern to applied.</returns>
    public static string GetShortDatePattern(TargetLocale locale)
    {
      DateTimeFormatInfo format = LocalizationUtil.GetDateFormat(locale);
      return format.ShortDatePattern;
    }

    /// <summary>
    /// Gets the short date pattern for a given locale which can
    /// be combined with other patterns (e.g. for formatting and
    /// parsing via the <see cref="FormatTimestamp"/> and
    /// <see cref="ParseTimestamp"/> methods.
    /// <para>
    /// CT's "short" date should not be confused with the .NET framework's
    /// short date. While the latter only contains hours and minutes, this
    /// pattern includes seconds as well.
    /// </para>
    /// </summary>
    /// <param name="locale">The targeted locale.</param>
    /// <returns>The date pattern to applied.</returns>
    public static string GetShortTimePattern(TargetLocale locale)
    {
      DateTimeFormatInfo format = LocalizationUtil.GetDateFormat(locale);
      return format.LongTimePattern;
    }

    #endregion


    #region write / parse any date based on formatting pattern

    /// <summary>
    /// Convenience method which formats a given timestamp with
    /// a formatting pattern for a given target locale.
    /// <para>
    /// The returned string can be parsed back into a matching
    /// DateTime value by invoking the <see cref="ParseTimestamp"/>
    /// methods with the same pattern and target locale.
    /// </para>
    /// </summary>
    /// <param name="dt">The date/time value to be formatted.</param>
    /// <param name="pattern">The format pattern to be applied.</param>
    /// <param name="locale">The targeted locale.</param>
    /// <returns>A string that represents the submitted timestamp according
    /// to the formatting pattern and the target locale.</returns>
    public static string FormatTimestamp(DateTime dt, string pattern, TargetLocale locale)
    {
      DateTimeFormatInfo format = LocalizationUtil.GetDateFormat(locale);
      return dt.ToString(pattern, format);
    }


    /// <summary>
    /// Converts a formatted timestamp string that was created by
    /// <see cref="FormatTimestamp"/> back into a <c>DateTime</c>
    /// struct.
    /// </summary>
    /// <param name="value">A string that represents a given time.</param>
    /// <param name="pattern">The format pattern that was applied to
    /// create the formatted string.</param>
    /// <param name="locale">The targeted locale.</param>
    /// <returns>A <c>DateTime</c> struct that represents the
    /// serialized timestamp.</returns>
    /// <remarks>
    /// The value parameter contains the date and time to parse.
    /// If the value parameter contains only a time and no date,
    /// the current date is used. If the value parameter contains
    /// only a date and no time, midnight (00:00:00) is used.
    /// The value parameter cannot contain leading, inner, or
    /// trailing white space characters.
    /// </remarks>
    public static DateTime ParseTimestamp(string value, string pattern, TargetLocale locale)
    {
      DateTimeFormatInfo format = LocalizationUtil.GetDateFormat(locale);
      return DateTime.ParseExact(value, pattern, format);
    }

    #endregion


    #region format time span

    /// <summary>
    /// Renders a timestamp depending on its actual range. This
    /// means that big ranges do not provide detailed fractions, while
    /// smaller ranges do not display days.
    /// </summary>
    /// <param name="range">The range to be displayed.</param>
    /// <param name="locale">Defines the culture to be used.</param>
    /// <returns></returns>
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA1305:SpecifyIFormatProvider", MessageId = "System.DateTime.ToString(System.String)")]
    public static string FormatTimeSpan(TimeSpan range, TargetLocale locale)
    {    
      StringBuilder builder = new StringBuilder();

      if (range.Ticks < 0)
      {
        //we have a negative range
        NumberFormatInfo numberFormat = LocalizationUtil.GetNumberFormat(locale);
        builder.Append(numberFormat.NegativeSign);
        range = range.Negate();
      }
      
      //render days if the range is that big
      if (range.Days > 0)
      {
        builder.Append(range.Days).Append(".");
      }

      //work with a DateTime struct which provides sophisticated formatting
      //(do not move above test for negative range!)
      DateTime timestamp = DateTime.MinValue.Add(range);
      
      //if the range is more than an hour, we're done
      if (range.TotalHours >= 24)
      {
        builder.Append(timestamp.ToString(ShortTimePattern));
        return builder.ToString();
      }

      string pattern;
      //determine how many fractions should be displayed
      if (range.TotalSeconds < 1)
      {
        pattern = MaxResolutionTimePattern;
      }
      else if (range.TotalMinutes < 1)
      {
        pattern = HighResolutionTimePattern;
      }
      else
      {
        pattern = DetailedTimePattern;
      }      

      //apply time pattern and return combined string
      builder.Append(timestamp.ToString(pattern));
      return builder.ToString();
    }
    
    #endregion

  }
}