using System;
using System.Diagnostics;
using System.Globalization;
using System.Threading;

namespace Hardcodet.Commons.Globalization
{
  /// <summary>
  /// Provides static util method which address common
  /// localization issues such as date / numeric formatting.
  /// </summary>
  /// <remarks>It does not make sense to move the formattign strings
  /// into localizable resource files. The application may rely
  /// on these patterns for serialization
  /// and deserialization cycles.</remarks>
  [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "Util"), DebuggerNonUserCode]
  public static class LocalizationUtil
  {
    #region get date / numeric formatters

    /// <summary>
    /// Gets a formatter for culture-aware or culture-independent
    /// formatting or parsing of date/time values.
    /// </summary>
    /// <param name="locale">Defines the source of the date format.</param>
    /// <returns>A matching date / time formatter. If the requested culture
    /// does not provide a number format (only if <paramref name="locale"/>
    /// is <see cref="TargetLocale.UI"/>, and the UI culture is neutral),
    /// a format for the OS installed culture is being returned.</returns>
    /// <exception cref="ArgumentOutOfRangeException">In case of an invalid localization
    /// flag.</exception>
    public static DateTimeFormatInfo GetDateFormat(TargetLocale locale)
    {
      switch (locale)
      {
        case TargetLocale.Invariant:
          return CultureInfo.InvariantCulture.DateTimeFormat;
        case TargetLocale.Thread:
          return Thread.CurrentThread.CurrentCulture.DateTimeFormat;
        case TargetLocale.UI:
          CultureInfo culture = Thread.CurrentThread.CurrentUICulture;
          if (culture.IsNeutralCulture)
          {
            //try fallback to specific culture
            culture = CultureInfo.InstalledUICulture;
          }
          return culture.DateTimeFormat;
      }

      throw new ArgumentOutOfRangeException("locale");
    }



    /// <summary>
    /// Gets a formatter for culture-aware or culture-independent
    /// formatting or parsing of numeric values.
    /// </summary>
    /// <param name="locale">Defines the source of the number format.</param>
    /// <returns>A matching numeric formatter. If the requested culture
    /// does not provide a number format (only if <paramref name="locale"/>
    /// is <see cref="TargetLocale.UI"/>, and the UI culture is neutral),
    /// a format for the OS installed culture is being returned.</returns>
    /// <exception cref="ArgumentOutOfRangeException">In case of an invalid localization
    /// flag.</exception>
    public static NumberFormatInfo GetNumberFormat(TargetLocale locale)
    {
      switch(locale)
      {
        case TargetLocale.Invariant:
          return CultureInfo.InvariantCulture.NumberFormat;
        case TargetLocale.Thread:
          return Thread.CurrentThread.CurrentCulture.NumberFormat;
        case TargetLocale.UI:
          CultureInfo culture = Thread.CurrentThread.CurrentUICulture;
          if (culture.IsNeutralCulture)
          {
            //fall back to installed culture
            culture = CultureInfo.InstalledUICulture;
          }
          return culture.NumberFormat;
          
      }

      throw new ArgumentOutOfRangeException("locale");
    }

    #endregion


    #region resolve culture

    /// <summary>
    /// Gets a culture for the current thread depending on the
    /// requested <paramref name="locale"/>.
    /// </summary>
    /// <param name="locale">Defines the targeted culture.</param>
    /// <returns>The matching <c>CultureInfo</c>, which also depends
    /// on the current thread.</returns>
    public static CultureInfo ResolveCulture(TargetLocale locale)
    {
      switch(locale)
      {
        case TargetLocale.Invariant:
          return CultureInfo.InvariantCulture;
        case TargetLocale.Thread:
          return CultureInfo.CurrentCulture;
        case TargetLocale.UI:
          return CultureInfo.CurrentUICulture;
        default:
          throw new ArgumentOutOfRangeException("locale");
      }
    }


    /// <summary>
    /// Gets a culture for a given thread depending on the
    /// requested <paramref name="locale"/>.
    /// </summary>
    /// <param name="locale">Defines the targeted culture.</param>
    /// <param name="thread">The thread to be processed.</param>
    /// <returns>The matching <c>CultureInfo</c>, depending on the
    /// submitted thread.</returns>
    public static CultureInfo ResolveCulture(TargetLocale locale, Thread thread)
    {
      switch (locale)
      {
        case TargetLocale.Invariant:
          return CultureInfo.InvariantCulture;
        case TargetLocale.Thread:
          return thread.CurrentCulture;
        case TargetLocale.UI:
          return thread.CurrentUICulture;
        default:
          throw new ArgumentOutOfRangeException("locale");
      }
    }

    #endregion
  }
}