﻿using System;
using System.IO;
using System.Reflection;

namespace Hardcodet.Commons.IO
{
  /// <summary>
  /// Util methods for file-related operations.
  /// </summary>
  public static class FileUtil
  {
    /// <summary>
    /// The location of the application's entry point, if any.
    /// </summary>
    /// <remarks>This is the path of the application executable, if there
    /// is any. In special environments (e.g. certain unit test runners),
    /// this path is not set at all.</remarks>
    private static readonly string entryPath = String.Empty;

    #region initialization

    /// <summary>
    /// Inits the helper class.
    /// </summary>
    static FileUtil()
    {
      //get the entry assembly
      var assembly = Assembly.GetEntryAssembly() ?? Assembly.GetExecutingAssembly();

      //do nothing if there is no entry assembly
      //(non-standard environment, e.g. unit tests)
      if (assembly != null)
      {
        entryPath = Path.GetDirectoryName(assembly.Location);
      }
    }

    #endregion


    #region get working folder, absolute/relative paths

    /// <summary>
    /// Gets the application working folder (usually the entry path
    /// of the application).
    /// </summary>
    public static string WorkingFolder
    {
      get { return entryPath ?? String.Empty; }
    }

    /// <summary>
    /// Gets an absolute path, based on the executable's location.
    /// </summary>
    /// <param name="filePath">A file path.</param>
    /// <returns>An absolute file path. If a relative path was submitted,
    /// an absolute path, based on the application's excecutable will be
    /// returned. If an absolute path was submitted, the unchanged path
    /// is being returned.</returns>
    public static string GetAbsolutePath(string filePath)
    {
      if (Path.IsPathRooted(filePath)) return filePath;

      //combine relative path with the location of the application's entry point
      return Path.Combine(entryPath, filePath).Replace("/", "\\");
    }


    /// <summary>
    /// Creates a relative path to a file regarding a 
    /// </summary>
    /// <param name="rootFolder">Root folder.</param>
    /// <param name="filePath">Path of the file.</param>
    /// <returns>Path to the file, if possible relative to the root folder.</returns>
    public static string GetRelativePath(string rootFolder, string filePath)
    {
      if (rootFolder == String.Empty) return filePath;

      //make sure we have a valid directory path
      rootFolder = rootFolder.Replace("\\", "/");
      if (!rootFolder.EndsWith("/")) rootFolder += "/";


      Uri baseUri = new Uri(rootFolder);
      Uri fileUri = new Uri(filePath);

      return baseUri.MakeRelativeUri(fileUri).ToString();
    }

    #endregion

    #region create temp file / folder

    /// <summary>
    /// Generates a path for a temporary file.
    /// </summary>
    /// <param name="prefix">The base name of the file.</param>
    /// <param name="extension">Extension of the file (without dot).</param>
    /// <returns></returns>
    /// <exception cref="ArgumentNullException">If <paramref name="prefix"/>
    /// is a null reference.</exception>
    /// <exception cref="ArgumentNullException">If <paramref name="extension"/>
    /// is a null reference.</exception>
    public static string CreateTempFilePath(string prefix, string extension)
    {
      if (prefix == null) throw new ArgumentNullException("prefix");
      if (extension == null) throw new ArgumentNullException("extension");
      string path = String.Format("{0}.{1}.{2}", prefix, Guid.NewGuid(), extension);
      return Path.Combine(Path.GetTempPath(), path);
    }

    /// <summary>
    /// Generates a path for a temporary file.
    /// </summary>
    /// <param name="rootFolder">The folder that contains the file.</param>
    /// <param name="prefix">The base name of the file.</param>
    /// <param name="extension">Extension of the file (without dot).</param>
    /// <returns></returns>
    /// <exception cref="ArgumentNullException">If <paramref name="prefix"/>
    /// is a null reference.</exception>
    /// <exception cref="ArgumentNullException">If <paramref name="extension"/>
    /// is a null reference.</exception>
    public static string CreateTempFilePath(string rootFolder, string prefix, string extension)
    {
      if (prefix == null) throw new ArgumentNullException("prefix");
      if (extension == null) throw new ArgumentNullException("extension");
      string path = String.Format("{0}.{1}.{2}", prefix, Guid.NewGuid(), extension);
      return Path.Combine(rootFolder, path);
    }


    /// <summary>
    /// Generates a path for a temporary file.
    /// </summary>
    /// <param name="extension">Extension of the file (without dot).</param>
    /// <returns></returns>
    /// <exception cref="ArgumentNullException">If <paramref name="extension"/>
    /// is a null reference.</exception>
    public static string CreateTempFilePath(string extension)
    {
      if (extension == null) throw new ArgumentNullException("extension");
      string path = String.Format("{0}.{1}", Guid.NewGuid(), extension);
      return Path.Combine(Path.GetTempPath(), path);
    }


    /// <summary>
    /// Creates a folder withing the system's temporary files folder
    /// that matches the submitted <paramref name="folderName"/>.
    /// If a folder with the submitted name already exists, a new
    /// folder with a numeric suffix (001, 002, ...) is being created.
    /// </summary>
    /// <param name="folderName">The suggested folder name.</param>
    public static DirectoryInfo CreateTempFolder(string folderName)
    {
      return CreateTempFolder(new DirectoryInfo(Path.GetTempPath()), folderName);
    }


    /// <summary>
    /// Creates a folder the at a given location that matches the submitted
    /// <paramref name="folderName"/>. If a folder with the submitted name
    /// already exists, a new folder with a numeric suffix (001, 002, ...)
    /// is being created.
    /// </summary>
    /// <param name="rootFolder">The root folder in which the temporary
    /// folder is being created.</param>
    /// <param name="folderName">The suggested folder name.</param>
    public static DirectoryInfo CreateTempFolder(DirectoryInfo rootFolder, string folderName)
    {
      var tempFolder = Path.Combine(rootFolder.FullName, folderName);

      //make sure the folder is new
      int fileCounter = 0;
      while (Directory.Exists(tempFolder))
      {
        //create a numeric suffix (001, 002, ...) until a unique folder name was found
        fileCounter++;
        tempFolder = Path.Combine(rootFolder.FullName, folderName + fileCounter.ToString("000"));
      }

      //create temp folder
      return Directory.CreateDirectory(tempFolder);
    }


    /// <summary>
    /// Ensures a unique file name by checking if a given file
    /// already exists, and returning the file name with a numeric
    /// suffix, if necessary.
    /// </summary>
    /// <param name="suggestedFileName">The suggested file name.</param>
    /// <returns>A qualified path to a file that does not exist yet.</returns>
    public static string GetUniqueFileName(string suggestedFileName)
    {
      string fileName = GetAbsolutePath(suggestedFileName);
      if (!File.Exists(fileName)) return fileName;

      string folder = Path.GetDirectoryName(fileName);
      string extension = Path.GetExtension(fileName);
      string strippedName = Path.GetFileNameWithoutExtension(fileName);

      int fileCounter = 0;
      while (File.Exists(fileName))
      {
        //create a numeric suffix (001, 002, ...) until a unique file name is found
        fileCounter++;
        fileName = Path.Combine(folder, strippedName + fileCounter.ToString("000") + extension);
      }

      return fileName;
    }

    #endregion

    #region copy directory

    /// <summary>
    /// Performs a recursive copy of a given directory.
    /// </summary>
    /// <param name="source"></param>
    /// <param name="destination"></param>
    /// <param name="overwrite"></param>
    public static void CopyDirectory(string source, string destination, bool overwrite)
    {
      // Create the destination folder if missing.
      if (!Directory.Exists(destination))
        Directory.CreateDirectory(destination);

      DirectoryInfo dirInfo = new DirectoryInfo(source);

      // Copy all files.
      foreach (FileInfo fileInfo in dirInfo.GetFiles())
        fileInfo.CopyTo(Path.Combine(destination, fileInfo.Name), overwrite);

      // Recursively copy all sub-directories.
      foreach (DirectoryInfo subDirectoryInfo in dirInfo.GetDirectories())
        CopyDirectory(subDirectoryInfo.FullName, Path.Combine(destination, subDirectoryInfo.Name), overwrite);
    }

    #endregion

    #region is directory read-only

    /// <summary>
    /// Checks whether a given folder is read-only or not.
    /// </summary>
    /// <param name="directory">The directory to be looked at.</param>
    /// <returns>True if the directory is readonly.</returns>
    /// <exception cref="ArgumentNullException">If <paramref name="directory"/>
    /// is a null reference.</exception>
    public static bool IsReadOnly(this DirectoryInfo directory)
    {
      if (directory == null) throw new ArgumentNullException("directory");
      var attributes = directory.Attributes;
      return (attributes & FileAttributes.ReadOnly) == FileAttributes.ReadOnly;
    }

    #endregion

    #region compute fingerprint

    /// <summary>
    /// Calculates a CRC code which can be used to ensure
    /// integrity of a given file.
    /// </summary>
    /// <param name="fileName">The file to be hashed.</param>
    /// <returns>A string representation of the file's CRC32
    /// hash code.</returns>
    /// <exception cref="ArgumentNullException">If <paramref name="fileName"/>
    /// is a null reference.</exception>
    /// <exception cref="FileNotFoundException">If the submitted
    /// file cannot be found.</exception>
    public static string ComputeFingerPrint(string fileName)
    {
      if (fileName == null) throw new ArgumentNullException("fileName");

      fileName = GetAbsolutePath(fileName);
      if (!File.Exists(fileName))
      {
        string msg = String.Format("File {0} does not exist.", fileName);
        throw new FileNotFoundException(msg);
      }

      return Crc32.ComputeFileHash(fileName).ToString();
    }


    /// <summary>
    /// Calculates a CRC code which can be used to ensure
    /// integrity of a given file.
    /// </summary>
    /// <param name="file">The file to be hashed.</param>
    /// <returns>A string representation of the file's CRC32
    /// hash code.</returns>
    /// <exception cref="ArgumentNullException">If <paramref name="file"/>
    /// is a null reference.</exception>
    /// <exception cref="FileNotFoundException">If the submitted
    /// file cannot be found.</exception>
    public static string ComputeFingerPrint(this FileInfo file)
    {
      if (file == null) throw new ArgumentNullException("file");
      return ComputeFingerPrint(file.FullName);
    }

    #endregion

  }
}