﻿using System;
using System.ComponentModel;

namespace Hardcodet.Commons
{
  /// <summary>
  /// A simple base class that provides an implementation of the <see cref="IDisposable"/>
  /// pattern on top of the <see cref="INotifyPropertyChanged"/> implementations of the
  /// <see cref="SimpleObject"/> base class.
  /// </summary>
  public class SimpleDisposableObject : SimpleObject, IDisposable
  {
    #region IsDisposed

    /// <summary>
    /// Tracks whether <see cref="Dispose()"/> has been called or not.
    /// </summary>
    private bool isDisposed;


    /// <summary>
    /// Tracks whether <see cref="Dispose()"/> has been called or not.
    /// </summary>
    public virtual bool IsDisposed
    {
      get { return isDisposed; }
      private set
      {
        isDisposed = value;
        OnPropertyChanged("IsDisposed");
      }
    }

    #endregion

    #region finalizer

    /// <summary>
    /// This destructor will run only if the <see cref="Dispose()"/>
    /// method does not get called. This gives this base class the
    /// opportunity to finalize.
    /// <para>
    /// Important: Do not provide destructors in types derived from
    /// this class.
    /// </para>
    /// </summary>
    ~SimpleDisposableObject()
    {
      //delegate disposal
      Dispose(false);
    }

    #endregion

    #region IDisposable pattern

    /// <summary>
    /// Disposes the object.
    /// </summary>
    /// <remarks>This method is not virtual by design. Derived classes
    /// should override <see cref="Dispose(bool)"/>.
    /// </remarks>
    public void Dispose()
    {
      Dispose(true);

      // This object will be cleaned up by the Dispose method.
      // Therefore, you should call GC.SupressFinalize to
      // take this object off the finalization queue 
      // and prevent finalization code for this object
      // from executing a second time.
      
      //only suppress in release builds
#if (!DEBUG)
      GC.SuppressFinalize(this);
#endif
    }


    /// <summary>
    /// <c>Dispose(bool disposing)</c> executes in two distinct scenarios.
    /// If disposing equals <c>true</c>, the method has been called directly
    /// or indirectly by a user's code. Managed and unmanaged resources
    /// can be disposed.<br/>
    /// Invoking this method sets the <see cref="IsDisposed" /> property. However,
    /// a property change event is only fired if <c>disposing</c> is true.
    /// </summary>
    /// <param name="disposing">If disposing equals <c>false</c>, the method
    /// has been called by the runtime from inside the finalizer and you
    /// should not reference other objects. Only unmanaged resources can
    /// be disposed.</param>
    /// <remarks>Check the <see cref="IsDisposed"/> property to determine whether
    /// the method has already been called.</remarks>
    protected virtual void Dispose(bool disposing)
    {
      if (isDisposed) return;

      if (disposing)
      {
        //fire property changed event
        IsDisposed = true;
      }
      else
      {
        isDisposed = true;
      }
    }

    #endregion
  }
}